<?php
  class pad_base {
    var $products_id;
    var $products_tax_class_id;
    var $show_out_of_stock;
    var $mark_out_of_stock;
    var $out_of_stock_msgline;
    var $no_add_out_of_stock;


/*
    Method: pad_base
    Class constructor
    Parameters:
      $products_id      integer     The product id of the product attributes are to be displayed for
    Returns:
      nothing
*/
    function pad_base($products_id = 0) {

      $this->products_id  = $products_id;

      if ($this->products_id != 0) {
        $tax_class_query_sql = 'SELECT products_tax_class_id
                                         FROM ' . TABLE_PRODUCTS . "
                                         WHERE products_id = '" . (int)$products_id . "'";
        $tax_class_query = tep_db_query($tax_class_query_sql);
        $tax_class_array = tep_db_fetch_array($tax_class_query);
        $this->products_tax_class_id = $tax_class_array['products_tax_class_id'];
      }

      $this->_SetConfigurationProperties('PRODINFO_ATTRIBUTE_');

    }


/*
    Method: _SetConfigurationProperties
    Set local configuration properties from osCommerce configuration DB constants
    Parameters:
      $prefix      sting     Prefix for the osCommerce DB constants
    Returns:
      nothing
*/
    function _SetConfigurationProperties($prefix) {

      $this->show_out_of_stock    = constant($prefix . 'SHOW_OUT_OF_STOCK');
      $this->mark_out_of_stock    = constant($prefix . 'MARK_OUT_OF_STOCK');
      $this->out_of_stock_msgline = constant($prefix . 'OUT_OF_STOCK_MSGLINE');
      $this->no_add_out_of_stock  = constant($prefix . 'NO_ADD_OUT_OF_STOCK');

    }


/*
    Method: draw

    Draws the product attributes.  This is the only method other than the constructor that is
    intended to be called by a user of this class.

    Attributes that stock is tracked for are grouped first and drawn with one dropdown list per
    attribute.  All attributes are drawn even if no stock is available for the attribute and no
    indication is given that the attribute is out of stock.

    Attributes that stock is not tracked for are then drawn with one dropdown list per
    attribute.

    Parameters:

      none

    Returns:

      string:       HTML for displaying the product attributes

*/
    function draw() {
      $out.=$this->_draw_stocked_attributes();
      $out.=$this->_draw_nonstocked_attributes();
      return $out;

    }

/*
    Method: _draw_stocked_attributes
    Draws the product attributes that stock is tracked for.
    Intended for class internal use only.

    Attributes that stock is tracked for are drawn with one dropdown list per attribute.
    All attributes are drawn even if no stock is available for the attribute and no
    indication is given that the attribute is out of stock.

    Parameters:
      none
    Returns:
      string:       HTML for displaying the product attributes that stock is tracked for

*/
    function _draw_stocked_attributes() {
      $out = '';

      $attributes = $this->_build_attributes_array(true, false);
      if (sizeof($attributes)>0) {
        foreach ($attributes as $stocked) {
          $out .= '<br /><strong>' . $stocked['oname'] . ":</strong>" . tep_draw_pull_down_menu('id['.$stocked['oid'].']',array_values($stocked['ovals']), $stocked['default'], ' class="productAttributes" id="id_attr_'.strtolower($stocked['oname']).'"') . "\n";
        }
      }
      return $out;
    }





	function _parse_all_attributes($products_id = 0){
		if( 0 == $products_id ) {
			return false;
		}

		$cpid = preg_split('/({\d{1,10}\})/',$products_id, -1, PREG_SPLIT_DELIM_CAPTURE);

		$attributes_array = array();

		for ( $i=0, $n = count($cpid); $i < $n; $i++ ) {
			if( 0 != $i ) {
				if( true == strstr($cpid[$i],'{') && true == strstr($cpid[$i],'}') ) {
					$attr_name = str_replace(array('{','}'), '', $cpid[$i]);
					$attr_val = $cpid[($i+1)];
					$attributes_array['id['.$attr_name.']'] = $attr_val;
				}
			}
		}
		return $attributes_array;
	}


/*
    Method: _draw_nonstocked_attributes
    Draws the product attributes that stock is not tracked for.
    Intended for class internal use only.
    Attributes that stock is not tracked for are drawn with one dropdown list per attribute.
    Parameters:
      none
    Returns:
      string:       HTML for displaying the product attributes that stock is not tracked for
*/

	function _draw_nonstocked_attributes() {

		$nonstocked_attributes = $this->_build_attributes_array(false, true);

    		foreach($nonstocked_attributes as $nonstocked) {
			switch ($nonstocked['type']){
				case 'dropdown':
					$output .= $this->_draw_dropdown_attributes($nonstocked);
				break;
				case 'radio':
					$output .= $this->_draw_radio_attributes($nonstocked);
				break;
				case 'checkbox':
					$output .= $this->_draw_checkbox_attributes($nonstocked);
				break;
			}
    		}

    		return $output;
    }


	function _draw_radio_attributes($nonstocked) {
		$selected_attributes = $this->_parse_all_attributes($this->products_id);
		$out .= '<p class="productAttributes"><span class="productAttributesTitle">'.$nonstocked['oname'].':</span><br />';
		foreach ( $nonstocked['ovals'] as $oval ) {
	        	if ( $nonstocked['default'] == $oval['id'] ) {
	        		$CHECKED = ' checked="checked"';
	        	} else {
	        		$CHECKED = ( $selected_attributes['id['.$nonstocked['oid'].']'] == $oval['id'] ? ' checked="checked"' : '' );
	        	}
	        	$input_id = 'id_'.$nonstocked['oid'] . '_' . $oval['id'];


	        	if ( $oval['image'] ) {
        			$oval['image'] = DIR_WS_IMAGES.$oval['image'];
        			###################   IMAGE SPECIAL MOD
				if ( 'on' == IMAGE_STAMP_ON_OFF ) {
					if( $oval['image_stamp'] ) {
						$oval['image'] = generate_image_stamp($oval['image'], DIR_WS_IMAGES.$oval['image_stamp']);
					}
				}
				###################   IMAGE SPECIAL MOD

				$img_width = LARGE_IMAGE_WIDTH;
				$img_height = LARGE_IMAGE_HEIGHT;

				if ( '' != $CHECKED ) {
					$javascript_selected_action = 'window.onload=function(){setAttributeProductImage'.$nonstocked['oid'].'('.$oval['id'].');}';
					}
					$javascript_inner_out .= ' if (selected_id == '.$oval['id'].') { '."\n";
					$javascript_inner_out .= ' var set_img = \'product_thumb.php?img='.$oval['image'].'&w='.$img_width.'&h='.$img_width.'\'; '."\n";
					$javascript_inner_out .= ' document.getElementById(\'href_prodimg\').href=\''.$oval['image'].'\';'."\n";
					$javascript_inner_out .= ' switchProdImage(set_img);'."\n";
					$javascript_inner_out .= ' } '."\n";
        		}
	        		$out .= '<input type="radio" name="id['.$nonstocked['oid'].']" value="'.$oval['id']. '" id="'.$input_id.'"'.$CHECKED . ' onclick="setAttributeProductImage'.$nonstocked['oid'].'(this.value);" /><label for="' . $input_id . '"> ' .
	        				$oval['text'] . '</label><br />';

        	}
        	$out .= stripslashes($nonstocked['odesc']) . '</p>';

        	if ( '' == $javascript_inner_out ) {
        	$javascript_inner_out = 'return false;';

        }

		$out .= '<script type="text/javascript">'."\n";
		$out .= "//<![CDATA[\n";
	        $out .= 'function setAttributeProductImage'.$nonstocked['oid'].'(selected_id) { '."\n";
	        $out .= $javascript_inner_out;
	        $out .= '}';
	        $out .= $javascript_selected_action;
	        $out .= "\n//]]>\n";
	        $out .= '</script>'."\n";

        $javascript_inner_out = '';
        $javascript_selected_action = '';

        return $out;
	}

	function _draw_dropdown_attributes($nonstocked) {

		$selected_attributes = $this->_parse_all_attributes($this->products_id);

		if ( 'true' == ATTRIBUTES_AUTO_PRICE_UPDATE ) {
			$js_update_price = "updatePrice('".$nonstocked['oid']."', this.options[this.selectedIndex].value);";
		}

        $out .= '<p class="productAttributes"><span class="productAttributesTitle">'.$nonstocked['oname'].':</span><br />';
        $out .= '<select name="id['.$nonstocked['oid'].']" class="productAttributes" onchange="javascript:'.$js_update_price.'setAttributeProductImage'.$nonstocked['oid'].'(this.options[this.selectedIndex].value);">';
        foreach ( $nonstocked['ovals'] as $oval ) {

        	if ( $nonstocked['default'] == $oval['id'] ) {
        		$SELECTED = ' selected="selected"';
        	} else {
        		$SELECTED = ( $selected_attributes['id['.$nonstocked['oid'].']'] == $oval['id'] ? ' selected="selected"' : '' );
        	}


        		if ( $oval['image'] ) {
        			$oval['image'] = DIR_WS_IMAGES.$oval['image'];
        			###################   IMAGE SPECIAL MOD
				if ( 'on' == IMAGE_STAMP_ON_OFF ) {
					if( $oval['image_stamp'] ) {
						$oval['image'] = generate_image_stamp($oval['image'], DIR_WS_IMAGES.$oval['image_stamp']);
					}
				}
				###################   IMAGE SPECIAL MOD

				$img_width = LARGE_IMAGE_WIDTH;
				$img_height = LARGE_IMAGE_HEIGHT;
				$oval['text'] = html_entity_decode($oval['text']);
				if ( '' != $SELECTED ) {
					$javascript_selected_action = 'window.onload=function(){setAttributeProductImage'.$nonstocked['oid'].'('.$oval['id'].');}';
					}
					$javascript_inner_out .= ' if (selected_id == '.$oval['id'].') { '."\n";
					$javascript_inner_out .= ' var set_img = \'product_thumb.php?img='.$oval['image'].'&w='.$img_width.'&h='.$img_width.'\'; '."\n";
					$javascript_inner_out .= ' document.getElementById(\'href_prodimg\').href=\''.$oval['image'].'\';'."\n";
					$javascript_inner_out .= ' switchProdImage(set_img);';
					$javascript_inner_out .= ' } '."\n";

					$out .= '<option value="'.$oval['id']."\"$SELECTED>".$oval['text'].'</option>';
        		} else {
        			$out .= '<option value="'.$oval['id']."\"$SELECTED>".$oval['text'].'</option>';
        		}
        }
        $out .= '</select> ';
        $out .= stripcslashes($nonstocked['odesc']);
        $out .= '</p>';
        if ( '' == $javascript_inner_out ) {
        	$javascript_inner_out = 'return false;';

        }

        $this->javascript .=
        '<script type="text/javascript">'."\n".
        "//<![CDATA[\n".
        'function setAttributeProductImage'.$nonstocked['oid'].'(selected_id) { '."\n".
        $javascript_inner_out.
        '}'.
        $javascript_selected_action.
        "\n//]]>\n".
        '</script>'."\n";

        $javascript_inner_out = '';
        $javascript_selected_action = '';

      return $out;

    }

    function _draw_checkbox_attributes($nonstocked) {
        $out .= '<p class="productAttributes"><span class="productAttributesTitle">'.$nonstocked['oname'].':</span><br />';
        foreach ( $nonstocked['ovals'] as $oval ) {
			$out .= "<input type=\"checkbox\" name=\"id[{$nonstocked['oid']}][]\" value=\"{$oval['id']}\" />{$oval['text']}<br />";
        }
        $out .= stripcslashes($nonstocked['odesc']);
        $out .= '</p>';
      return $out;
    }




/*
    Method: _build_attributes_array

    Build an array of the attributes for the product

    Parameters:

      $build_stocked        boolean   Flag indicating if stocked attributes should be built.
      $build_nonstocked     boolean   Flag indicating if non-stocked attribute should be built.

    Returns:

      array:                Array of attributes for the product of the form:
                              'oid'       => integer: products_options_id
                              'oname'     => string:  products_options_name
                              'ovals'     => array:   option values for the option id of the form
                                             'id'    => integer:  products_options_values_id
                                             'text'  => string:   products_options_values_name
                              'default'   => integer: products_options_values_id that the product id
                                                      contains for this option id and should be the
                                                      default selection when this attribute is drawn.
                                                      Set to zero if the product id did not contain
                                                      this option.

*/
    function _build_attributes_array($build_stocked, $build_nonstocked) {
      global $languages_id;
      global $currencies;
      global $cart;

      if (!($build_stocked | $build_nonstocked)) return null;

      if ($build_stocked && $build_nonstocked) {
        $stocked_where='';
      }
      elseif ($build_stocked) {
        $stocked_where="and popt.products_options_track_stock = '1'";
      }
      elseif ($build_nonstocked) {
        $stocked_where="and popt.products_options_track_stock = '0'";
      }
		$products_options_name_query_sql = "
									SELECT DISTINCT
										popt.products_options_id,
										poptd.products_options_name,
										poptd.products_options_description,
										popt.products_options_track_stock,
										popt.type
									FROM
										" . TABLE_PRODUCTS_OPTIONS . " popt,
										" . TABLE_PRODUCTS_OPTIONS_DESCRIPTION . " poptd,
										" . TABLE_PRODUCTS_ATTRIBUTES . " patrib
									WHERE
										patrib.products_id='" . (int)$this->products_id . "'
										AND patrib.options_id = popt.products_options_id
										AND popt.products_options_id = poptd.products_options_id
										AND poptd.language_id = '" . (int)$languages_id . "'
										" . $stocked_where . "
									ORDER BY
										popt.products_options_sort_order";
		$products_options_name_query = tep_db_query($products_options_name_query_sql);
		$attributes = array();
		while ( $products_options_name = tep_db_fetch_array($products_options_name_query) ) {
        $products_options_array = array();
        $products_options_query = tep_db_query("select pov.products_options_values_id, pov.products_options_values_name, pa.options_values_price, pa.price_prefix, pa.selected, pa.products_attributes_image, pa.products_attributes_image_stamp from " . TABLE_PRODUCTS_ATTRIBUTES . " pa, " . TABLE_PRODUCTS_OPTIONS_VALUES . " pov where pa.products_id = '" . (int)$this->products_id . "' and pa.options_id = '" . (int)$products_options_name['products_options_id'] . "' and pa.options_values_id = pov.products_options_values_id and pov.language_id = '" . (int)$languages_id . "' order by pa.products_options_sort_order");
        while ($products_options = tep_db_fetch_array($products_options_query)) {
          $products_options_array[] = array('id' => $products_options['products_options_values_id'], 'text' => 	htmlentities($products_options['products_options_values_name']), 'image' => $products_options['products_attributes_image'],'image_stamp' => $products_options['products_attributes_image_stamp']);
          if ($products_options['options_values_price'] != '0') {
          	$products_options_array[sizeof($products_options_array)-1]['text'] .= ' (' . $products_options['price_prefix'] . $currencies->display_price($products_options['options_values_price'], tep_get_tax_rate($this->products_tax_class_id)) .')';
          }
		  		 if($products_options['selected']=='true') $preselected = $products_options['products_options_values_id'];
        }
        if (isset($cart->contents[$this->products_id]['attributes'][$products_options_name['products_options_id']]))
          $selected = $cart->contents[$this->products_id]['attributes'][$products_options_name['products_options_id']];
		  else if($preselected)
		  $selected = $preselected;
		  else
          $selected = 0;
        $attributes[]=array('oid'=>$products_options_name['products_options_id'],
                            'oname'=>$products_options_name['products_options_name'],
                            'odesc'=>$products_options_name['products_options_description'],
                            'ovals'=>$products_options_array,
                            'type'=>$products_options_name['type'],
                            'default'=>$selected);
      }
      return $attributes;
    }


/*
    Method: _build_attributes_combinations

    A recursive method for building an array enumerating the attribute combinations for the product

    Parameters:

      $attributes             array     An array of the attributes that combinations will be built for.
                                        Format is as returned by _build_attributes_array.
      $showoos                boolean   Flag indicating if non-stocked attributes should be built.
      $markoos                string    'Left' if out of stock indication is to be appended in front of the
                                        attribute combination text.  'Right' if out of stock indication is
                                        to be appended at the end of the attribute combination text.
      $combinations           array     Array of the attribute combinations is returned in this parameter.
                                        Should be set to an empty array before an external call to this method.
                                          'comb'        => array:   array of a single attribute combination
                                                                      options_id => options_value_id
                                          'id'          => string:  options/values string for this
                                                                    combination in the form for the
                                                                    key of the products_stock table
                                                                      opt_id-val_id,opt_id-val_id,...
                                          'text'        => string:  Text for this combination.  Values text
                                                                    is as built by _build_attributes_array
                                                                     and contains the add/subtract price for
                                                                     the option value if applicable.  Form is:
                                                                       values_text, values_text
      $selected_combination   integer   Index into the $combinations array of the combination that should
                                        be the default selection when the combination is drawn is returned in
                                        this parameter.  Determined from product id.  Should be set to zero
                                        before an external call to this method.

    Parameters for internal recursion use only:

      $oidindex               integer   Index into the $attributes array of the option to operate on.
      $comb                   array     Array containing option id/values of combination built so far
                                          products_options_id => products_options_value_id
      $id                     string    Contains string of options/values built so far
      $text                   string    Text for the options values constructed so far.
      $isselected             boolean   Flag indicating if so far all option values in this combination
                                        were indicated to be defaults in the product id.


    Returns:

      see $combinations and $selected_combination parameters above
      no actual function return value.

*/
    function _build_attributes_combinations($attributes, $showoos, $markoos, &$combinations, &$selected_combination, $oidindex=0, $comb=array(), $id="", $text='', $isselected=true) {
      global $cart;

      foreach ($attributes[$oidindex]['ovals'] as $attrib) {
        $newcomb = $comb;
        $newcomb[$attributes[$oidindex]['oid']] = $attrib['id'];
        $newid=$id.','.$attributes[$oidindex]['oid'].'-'.$attrib['id'];
        $newtext = $text.", ".$attrib['text'];
        if (isset($cart->contents[$this->products_id]['attributes'][$attributes[$oidindex]['oid']]))
          $newisselected = ($cart->contents[$this->products_id]['attributes'][$attributes[$oidindex]['oid']] == $attrib['id']) ? $isselected : false;
        else
          $newisselected = false;
        if (isset($attributes[$oidindex+1])) {
          $this->_build_attributes_combinations($attributes, $showoos, $markoos, $combinations, $selected_combination, $oidindex+1, $newcomb, $newid, $newtext, $newisselected);
        }
        else {
          $is_out_of_stock=tep_check_stock(tep_get_prid($this->products_id),1,$newcomb);
          if (!$is_out_of_stock | ($showoos == true)) {
            switch ($markoos) {
              case 'Left':   $newtext=($is_out_of_stock ? TEXT_OUT_OF_STOCK.' - ' : '').substr($newtext,2);
                             break;
              case 'Right':  $newtext=substr($newtext,2).($is_out_of_stock ? ' - '.TEXT_OUT_OF_STOCK : '');
                             break;
              default:       $newtext=substr($newtext,2);
                             break;
            }
            $combinations[] = array('comb'=>$newcomb, 'id'=>substr($newid,1), 'text'=>$newtext);
            if ($newisselected) $selected_combination = sizeof($combinations)-1;
          }
        }
      }
    }


/*
    Method: _draw_js_stock_array

    Draw a Javascript array containing the given attribute combinations.
    Generally used to draw array of in-stock combinations for Javascript out of stock
    validation and messaging.

    Parameters:

      $combinations        array   Array of combinations to build the Javascript array for.
                                   Array must be of the form returned by _build_attributes_combinations
                                   Usually this array only contains in-stock combinations.

    Returns:

      string:                 Javacript array definition.  Excludes the "var xxx=" and terminating ";".  Form is:
                              {optval1:{optval2:{optval3:1,optval3:1}, optval2:{optval3:1}}, optval1:{optval2:{optval3:1}}}
                              For example if there are 3 options and the instock value combinations are:
                                opt1   opt2   opt3
                                  1      5      4
                                  1      5      8
                                  1     10      4
                                  3      5      8
                              The string returned would be
                                {1:{5:{4:1,8:1}, 10:{4:1}}, 3:{5:{8:1}}}

*/
    function _draw_js_stock_array($combinations) {
      if (!((isset($combinations)) && (is_array($combinations)) && (sizeof($combinations) > 0))){
        return '{}';
      }
      $out='';
      foreach ($combinations[0]['comb'] as $oid=>$ovid) {
        $out.='{'.$ovid.':';
        $opts[]=$oid;
      }
      $out.='1';

      for ($combindex=1; $combindex<sizeof($combinations); $combindex++) {
        $comb=$combinations[$combindex]['comb'];
        for ($i=0; $i<sizeof($opts)-1; $i++) {
          if ($comb[$opts[$i]]!=$combinations[$combindex-1]['comb'][$opts[$i]]) break;
        }
        $out.=str_repeat('}',sizeof($opts)-1-$i).',';
        if ($i<sizeof($opts)-1) {
          for ($j=$i; $j<sizeof($opts)-1; $j++)
            $out.=$comb[$opts[$j]].':{';
        }
        $out.=$comb[$opts[sizeof($opts)-1]].':1';
      }
      $out.=str_repeat('}',sizeof($opts));

      return $out;
    }

  }
?>
