<?php


	function fastimagecopyresampled (&$dst_image, $src_image, $dst_x, $dst_y, $src_x, $src_y, $dst_w, $dst_h, $src_w, $src_h, $quality = 3) {
	  // Plug-and-Play fastimagecopyresampled function replaces much slower imagecopyresampled.
	  // Just include this function and change all "imagecopyresampled" references to "fastimagecopyresampled".
	  // Typically from 30 to 60 times faster when reducing high resolution images down to thumbnail size using the default quality setting.
	  // Author: Tim Eckel - Date: 12/17/04 - Project: FreeRingers.net - Freely distributable.
	  //
	  // Optional "quality" parameter (defaults is 3).  Fractional values are allowed, for example 1.5.
	  // 1 = Up to 600 times faster.  Poor results, just uses imagecopyresized but removes black edges.
	  // 2 = Up to 95 times faster.  Images may appear too sharp, some people may prefer it.
	  // 3 = Up to 60 times faster.  Will give high quality smooth results very close to imagecopyresampled.
	  // 4 = Up to 25 times faster.  Almost identical to imagecopyresampled for most images.
	  // 5 = No speedup.  Just uses imagecopyresampled, highest quality but no advantage over imagecopyresampled.
	  if (empty($src_image) || empty($dst_image)) { return false; }
	  if ($quality <= 1) {
	    $temp = imagecreatetruecolor ($dst_w + 1, $dst_h + 1);
	    imagecopyresized ($temp, $src_image, $dst_x, $dst_y, $src_x, $src_y, $dst_w + 1, $dst_h + 1, $src_w, $src_h);
	    imagecopyresized ($dst_image, $temp, 0, 0, 0, 0, $dst_w, $dst_h, $dst_w, $dst_h);
	    imagedestroy ($temp);
	  } elseif ($quality < 5 && (($dst_w * $quality) < $src_w || ($dst_h * $quality) < $src_h)) {
	    $tmp_w = $dst_w * $quality;
	    $tmp_h = $dst_h * $quality;
	    $temp = imagecreatetruecolor ($tmp_w + 1, $tmp_h + 1);
	    imagecopyresized ($temp, $src_image, $dst_x * $quality, $dst_y * $quality, $src_x, $src_y, $tmp_w + 1, $tmp_h + 1, $src_w, $src_h);
	    imagecopyresampled ($dst_image, $temp, 0, 0, 0, 0, $dst_w, $dst_h, $tmp_w, $tmp_h);
	    imagedestroy ($temp);
	  } else {
	    imagecopyresampled ($dst_image, $src_image, $dst_x, $dst_y, $src_x, $src_y, $dst_w, $dst_h, $src_w, $src_h);
	  }
	  return true;
	}


	function outputErrorImage($message){
		$im  = imagecreatetruecolor(250, 100);
		$bgc = imagecolorallocate($im, 255, 200, 200);
		$tc  = imagecolorallocate($im, 0, 0, 0);
		imagefilledrectangle($im, 0, 0, 250, 100, $bgc);
		imagestring($im, 2, 5, 5, $message, $tc);
		header("Content-Type: image/png");
		imagepng($im);
	}


	function generate_imagestamped_name($image, $stamp_image, $stamp_image_layering = 'WT') {
		if ( true === empty($image) ) {
			return false;
		}
		if( false === defined('IMG_STAMP_FILENAME_PREFIX') ) {
			define('IMG_STAMP_FILENAME_PREFIX','img_pr_');
		}
		list($image_base_name, $image_base_ext) = explode('.', basename($image));

		$return_filename = '';

		if('.' == dirname($image) ) {
				$return_filename .= IMG_STAMP_FILENAME_PREFIX;
		} else {
				$return_filename .= dirname($image) . '/' . IMG_STAMP_FILENAME_PREFIX;
		}

		if ( 'WT' === $stamp_image_layering ) { # watermark on top
			 $return_filename .= $image_base_name . '.jpg';
		} else { # watermark under
			if ( false === empty($stamp_image) ) {
				$stamp_image = basename($stamp_image);
			}
			list($stmp_img, $stmp_ext) = explode('.', $stamp_image);
			$return_filename .= $image_base_name . $stmp_img. '.jpg';
		}
		return $return_filename;

	}


	function watermark_image($main_image_file, $watermark_file, $stamp_image_options = ''){

		if ( (false == is_file($main_image_file) || false == is_readable($main_image_file)) ) {
			$message = "Error: Invalid main image";
			outputErrorImage($message);
		} else if ( (false == is_file($watermark_file) || false == is_readable($watermark_file)) ) {
			$message = "Error: Invalid watermark image";
			outputErrorImage($message);
		}


		## stamp_img options mod
		# watermark_1.png|P-BR:L-WT

		/*
	     	-----------------
	     	P positions:
	     	-----------------
	     	P-RC right center
	     	P-RT right top
	     	P-RB right bottom
	     	P-LC left center
	     	P-LT left top
	     	P-LB left bottom
	     	P-CT center top
	     	P-CB center bottom
	     	P-AC absolute center
	     	-----------------

	     	-----------------
	     	L layer:
	     	-----------------
			L-WT watermark on top
			L-WU watermark under
	     	-----------------

	     	-----------------
	     	PX-10 position-x:
	     	PY-134 position-y:
	     	-----------------
			integer or %
	     	-----------------

	     	-----------------
	     	Q-90 image quality:
	     	-----------------
			from 1 to 100
	     	-----------------
	     */

		$stamp_image_position = 'CA';
		$stamp_image_layering = 'WT';
		$image_quality = 90;

		if ( false == empty($stamp_image_options) ) {
			$options = explode(':', $stamp_image_options);
			if( true === is_array($options) ) {
				foreach ( $options as $option ) {
					list($opt, $val) = explode('-', $option);
					switch($opt) {
						case 'P':
							#position
							$stamp_image_position = $val;
						break;

						case 'L':
							#layer
							$stamp_image_layering = $val;
						break;

						case 'PX':
							#position-x
							$stamp_image_position_x = $val;
						break;

						case 'PY':
							#position-y
							$stamp_image_position_y = $val;

						break;

						case 'Q':
							#position-y
							$image_quality = $val;
						break;
					}
				}
			}

		}

		list($image_name, $image_ext) = explode('.', basename($main_image_file));
		$image_ext = strtolower($image_ext);

		list($watermark_name, $watermark_ext) = explode('.', basename($watermark_file));
		$watermark_ext = strtolower($watermark_ext);


		############## WATERMARK LAYER
		############## WT watermark on top
		if ( 'WT' == $stamp_image_layering ) {
		############## WATERMARK LAYER

			switch ( strtolower($watermark_ext) ) {
				case 'png':
					$watermark = imagecreatefrompng($watermark_file);
				break;

				case 'jpg':
				case 'jpeg':
					$watermark = imagecreatefromjpeg($watermark_file);
				break;

				case 'gif':
					$watermark = imagecreatefromgif($watermark_file);
				break;

				default:
					return false;
				break;
			}

			switch ( strtolower($image_ext) ) {
				case 'png':
				     $sz = getimagesize($main_image_file);
					$image_png = imagecreatefrompng($main_image_file);
					$image = imagecreatetruecolor($sz[0], $sz[1]);
					$bgcolor = imagecolorallocate($image, 255, 255, 255);
					imagefill($image , 0, 0 , $bgcolor);
				     imagecopy($image, $image_png, 0, 0, 0, 0, $sz[0], $sz[1]);
				     imagedestroy($image_png);
				break;

				case 'jpg':
				case 'jpeg':
					$image = imagecreatefromjpeg($main_image_file);
				break;

				case 'gif':
					$image = imagecreatefromgif($main_image_file);
				break;

				default:
					return false;
				break;
			}




		$image_src_x = imagesx($image);
	     $image_src_y = imagesy($image);
	     $watermark_src_x = imagesx($watermark);
	     $watermark_src_y = imagesy($watermark);

		$image_src_size_x = $watermark_src_x;
		$image_src_size_y = $watermark_src_y;

		$image_dst_size_x = $watermark_src_x;
		$image_dst_size_y = $watermark_src_y;

		############## WATERMARK LAYER
		############## WT watermark under
		} else {
		############## WATERMARK LAYER

		switch ( strtolower($image_ext) ) {
				case 'png':

					$watermark = imagecreatefrompng($main_image_file);
				break;

				case 'jpg':
				case 'jpeg':
					$watermark = imagecreatefromjpeg($main_image_file);
				break;

				case 'gif':
					$watermark = imagecreatefromgif($main_image_file);
				break;

				default:
					return false;
				break;
			}

		switch ( strtolower($watermark_ext) ) {
				case 'png':

					$sz = getimagesize($watermark_file);
					$image_png = imagecreatefrompng($watermark_file);
					$image = imagecreatetruecolor($sz[0], $sz[1]);
					$bgcolor = imagecolorallocate($image, 255, 255, 255);
					imagefill($image , 0, 0 , $bgcolor);
				     imagecopy($image, $image_png, 0, 0, 0, 0, $sz[0], $sz[1]);
				     imagedestroy($image_png);
				break;

				case 'jpg':
				case 'jpeg':
					$image = imagecreatefromjpeg($watermark_file);
				break;

				case 'gif':
					$image = imagecreatefromgif($watermark_file);
				break;

				default:
					return false;
				break;
			}


		$image_src_x = imagesx($image);
	     $image_src_y = imagesy($image);
	     $watermark_src_x = imagesx($watermark);
	     $watermark_src_y = imagesy($watermark);

		$image_src_size_x = $watermark_src_x;
		$image_src_size_y = $watermark_src_y;

		$image_dst_size_x = $watermark_src_x;
		$image_dst_size_y = $watermark_src_y;

		############## WATERMARK LAYER
		}
		############## WATERMARK LAYER

		if ( false === empty($stamp_image_position_x) && false === empty($stamp_image_position_y) ) {

			if( '%' == substr($stamp_image_position_x, -1) ) {
				$xpos = substr($stamp_image_position_x, 0, -1);

				$watermark_position_x =  round((($image_src_x * $xpos) / 100), 0);
				$watermark_position_x = ($watermark_position_x - ($watermark_src_x/2));

				if ( $watermark_position_x > ($image_src_x - $watermark_src_x) ) {
					$watermark_position_x = ($image_src_x - $watermark_src_x);
				} else if ( $watermark_position_x < ($watermark_src_x-($watermark_src_x/2)) ) {
					$watermark_position_x = 0;
				}

			} else {
				$watermark_position_x = $stamp_image_position_x;
			}

			if( '%' == substr($stamp_image_position_y, -1) ) {
				$ypos = substr($stamp_image_position_y, 0, -1);
				$watermark_position_y =  round((($image_src_y * $ypos) / 100), 0);
				$watermark_position_y = ($watermark_position_y - ($watermark_src_y/2));

				if ( $watermark_position_y > ($image_src_y - $watermark_src_y) ) {
					$watermark_position_y = ($image_src_y - $watermark_src_y);
				} else if ( $watermark_position_y < ($watermark_src_y-($watermark_src_y/2)) ) {
					$watermark_position_y = 0;
				}

			} else {
				$watermark_position_y = $stamp_image_position_y;
			}

		} else if ( $image_src_x > $watermark_src_x || $image_src_y > $watermark_src_y ) {

			switch ( $stamp_image_position ) {
			     case 'RC': #RIGHT_CENTER
			     	$watermark_position_x = ($image_src_x - $watermark_src_x);
					$watermark_position_y = (($image_src_y/2) - ($watermark_src_y/2));
			     break;

			     case 'RT': #RIGHT_TOP
			     	$watermark_position_x = ($image_src_x - $watermark_src_x);
					$watermark_position_y = 0;
			     break;

			     case 'RB': #RIGHT_BOTTOM
					$watermark_position_x = ($image_src_x - $watermark_src_x);
					$watermark_position_y = ($image_src_y - $watermark_src_y);

			     break;

			     case 'LC': #LEFT_CENTER
					$watermark_position_x = 0;
					$watermark_position_y = (($image_src_y/2) - ($watermark_src_y/2));
			     break;

			     case 'LT': #LEFT_TOP
				     $watermark_position_x = 0;
					$watermark_position_y = 0;
			     break;

			     case 'LB': #LEFT_BOTTOM
			     	$watermark_position_x = 0;
					$watermark_position_y = ($image_src_y - $watermark_src_y);
			     break;

			     case 'CT': #CENTER_TOP
					$watermark_position_x = ($image_src_x/2 - ( $watermark_src_x / 2));
					$watermark_position_y = 0;
			     break;

			     case 'CB': #CENTER_BOTTOM
					$watermark_position_x = (($image_src_x/2) - ($watermark_src_x/2));
					$watermark_position_y = ($image_src_y - $watermark_src_y);
			     break;

			     case 'CA': #CENTER ABSOLUTE
			     default:
				     $watermark_position_x = (($image_src_x/2) - ($watermark_src_x/2));
					$watermark_position_y = (($image_src_y/2) - ($watermark_src_y/2));
			     break;
			}

		} else {
			$watermark_position_x = 0;
			$watermark_position_y = 0;
		}

		$new_file = generate_imagestamped_name($main_image_file, $watermark_file, $stamp_image_layering);

		fastimagecopyresampled($image, $watermark, $watermark_position_x, $watermark_position_y, 0, 0, $image_src_size_x, $image_src_size_y, $image_dst_size_x, $image_dst_size_y);
		imagejpeg($image, $new_file, $image_quality);
		imagedestroy($image);

		}



	function generate_image_stamp($main_image, $stamp_image) {

		list($stamp_image, $stamp_image_options) = explode('|', $stamp_image);
		// cannot locate main image or warermark.

		if ( false === file_exists($main_image) || false === file_exists($stamp_image) ) {
			return $main_image;
		}
			//$stamp_image_options
			$stamp_image_layering = '';

			### Get watermark layering option value - to pas it to: generate_imagestamped_name
			$options = explode(':', $stamp_image_options);
			if( true === is_array($options) ) {
				foreach ( $options as $option ) {
					list($opt, $val) = explode('-', $option);
					if ( 'L' == $opt ) { #layer
						$stamp_image_layering = $val;
					}
				}
			}
			### ~
			$filename_img_out = generate_imagestamped_name($main_image, $stamp_image, $stamp_image_layering);

		if ( true === file_exists($filename_img_out) ) {

			if( filemtime($stamp_image) > filemtime($filename_img_out) || filemtime($main_image) > filemtime($filename_img_out) ) {
				//new file created, something was modified.
				watermark_image($main_image, $stamp_image, $stamp_image_options);
				return $filename_img_out;
			} else {
				//file found, nothing was modified, using old file.
				return $filename_img_out;
			}
		} else {
			//File does not exist - will create one.
			watermark_image($main_image, $stamp_image, $stamp_image_options);
			return $filename_img_out;
		}
		/**/
	}

?>