<?php
/**
* Encode hash
*/
function _hash_encode64($input, $count, &$itoa64)
{
	$output = '';
	$i = 0;

	do
	{
		$value = ord($input[$i++]);
		$output .= $itoa64[$value & 0x3f];

		if ($i < $count)
		{
			$value |= ord($input[$i]) << 8;
		}

		$output .= $itoa64[($value >> 6) & 0x3f];

		if ($i++ >= $count)
		{
			break;
		}

		if ($i < $count)
		{
			$value |= ord($input[$i]) << 16;
		}

		$output .= $itoa64[($value >> 12) & 0x3f];

		if ($i++ >= $count)
		{
			break;
		}

		$output .= $itoa64[($value >> 18) & 0x3f];
	}
	while ($i < $count);

	return $output;
}
/**
* Generate salt for hash generation
*/
function _hash_gensalt_private($input, &$itoa64, $iteration_count_log2 = 6)
{
	if ($iteration_count_log2 < 4 || $iteration_count_log2 > 31)
	{
		$iteration_count_log2 = 8;
	}

	$output = '$H$';
	$output .= $itoa64[min($iteration_count_log2 + ((PHP_VERSION >= 5) ? 5 : 3), 30)];
	$output .= _hash_encode64($input, 6, $itoa64);

	return $output;
}

/**
* The crypt function/replacement
*/
function _hash_crypt_private($password, $setting, &$itoa64)
{
	$output = '*';

	// Check for correct hash
	if (substr($setting, 0, 3) != '$H$')
	{
		return $output;
	}

	$count_log2 = strpos($itoa64, $setting[3]);

	if ($count_log2 < 7 || $count_log2 > 30)
	{
		return $output;
	}

	$count = 1 << $count_log2;
	$salt = substr($setting, 4, 8);

	if (strlen($salt) != 8)
	{
		return $output;
	}

	/**
	* We're kind of forced to use MD5 here since it's the only
	* cryptographic primitive available in all versions of PHP
	* currently in use.  To implement our own low-level crypto
	* in PHP would result in much worse performance and
	* consequently in lower iteration counts and hashes that are
	* quicker to crack (by non-PHP code).
	*/
	if (PHP_VERSION >= 5)
	{
		$hash = md5($salt . $password, true);
		do
		{
			$hash = md5($hash . $password, true);
		}
		while (--$count);
	}
	else
	{
		$hash = pack('H*', md5($salt . $password));
		do
		{
			$hash = pack('H*', md5($hash . $password));
		}
		while (--$count);
	}

	$output = substr($setting, 0, 12);
	$output .= _hash_encode64($hash, 16, $itoa64);

	return $output;
}

/**
* Check for correct password
*/
function phpbb_check_hash($password, $hash)
{
	$itoa64 = './0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz';
	if (strlen($hash) == 34)
	{
		return (_hash_crypt_private($password, $hash, $itoa64) === $hash) ? true : false;
	}

	return (md5($password) === $hash) ? true : false;
}

function phpbb_hash($password)
{
	$itoa64 = './0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz';

	$random_state = unique_id();
	$random = '';
	$count = 6;

	if (($fh = @fopen('/dev/urandom', 'rb')))
	{
		$random = fread($fh, $count);
		fclose($fh);
	}

	if (strlen($random) < $count)
	{
		$random = '';

		for ($i = 0; $i < $count; $i += 16)
		{
			$random_state = md5(unique_id() . $random_state);
			$random .= pack('H*', md5($random_state));
		}
		$random = substr($random, 0, $count);
	}

	$hash = _hash_crypt_private($password, _hash_gensalt_private($random, $itoa64), $itoa64);

	if (strlen($hash) == 34)
	{
		return $hash;
	}

	return md5($password);
}

function unique_id($extra = 'c')
{
	static $dss_seeded = false;
	global $config;

	$val = $config['rand_seed'] . microtime();
	$val = md5($val);
	$config['rand_seed'] = md5($config['rand_seed'] . $val . $extra);

	return substr($val, 4, 16);
}

function phpbb_get_activation_key($username) {
	if ( true === empty($username) ) {
		return false;
	}
	$phpbb_activation_key_query_sql = "SELECT user_actkey FROM " . PHPBB_DB_PREFIX . "users WHERE username_clean = '$username'";
	$phpbb_activation_key_query = tep_db_query($phpbb_activation_key_query_sql);
	if ( 0 < tep_db_num_rows($phpbb_activation_key_query) ) {
		$phpbb_activation_key = tep_db_fetch_array($phpbb_activation_key_query);
		return $phpbb_activation_key['user_actkey'];
	} else {
		return false;
	}
}

function phpbb_get_user_id($username) {
	if ( true === empty($username) ) {
		return false;
	}
	$phpbb_user_id_query_sql = "SELECT user_id FROM " . PHPBB_DB_PREFIX . "users WHERE username_clean = '$username'";
	$phpbb_user_id_query = tep_db_query($phpbb_user_id_query_sql);
	if ( 0 < tep_db_num_rows($phpbb_user_id_query) ) {
		$phpbb_user_id = tep_db_fetch_array($phpbb_user_id_query);
		return $phpbb_user_id['user_id'];
	} else {
		return false;
	}
}

function phpbb_username_exists($username){
	if ( true === empty($username) ) {
		return false;
	}
	$phpbb_user_query_sql = "SELECT user_id FROM " . PHPBB_DB_PREFIX . "users WHERE username = '$username'";
	$phpbb_user_query = tep_db_query($phpbb_user_query_sql);
	if ( 0 < tep_db_num_rows($phpbb_user_query) ) {
		return true;
	} else {
		return false;
	}
}
function phpbb_email_exists($email){
	if ( true === empty($email) ) {
		return false;
	}
	$phpbb_user_email_query_sql = "SELECT user_id FROM " . PHPBB_DB_PREFIX . "users WHERE user_email = '$email'";
	$phpbb_user_email_query = tep_db_query($phpbb_user_email_query_sql);
	if ( 0 < tep_db_num_rows($phpbb_user_email_query) ) {
		return true;
	} else {
		return false;
	}
}


function phpbb_get_user_id_by_customers_id($customers_id){
	if ( true === empty($customers_id) ) {
		return false;
	}
 	$phpbb_customers_id_query_sql = "SELECT customers_email_address, customers_user_id FROM " . TABLE_CUSTOMERS . " WHERE `customers_id` = '$customers_id'";
 	$phpbb_customers_id_query = tep_db_query($phpbb_customers_id_query_sql);
 	if ( 0 < tep_db_num_rows($phpbb_customers_id_query) ) {
 		$phpbb_customers_id = tep_db_fetch_array($phpbb_customers_id_query);

 		$phpbb_userid_query_sql = "SELECT user_id, username_clean FROM " . PHPBB_DB_PREFIX . "users WHERE (`username_clean` = '{$phpbb_customers_id['customers_user_id']}' OR `user_email` = '{$phpbb_customers_id['customers_email_address']}')";
 		$phpbb_userid_query = tep_db_query($phpbb_userid_query_sql);
 		if ( 0 < tep_db_num_rows($phpbb_userid_query) ) {
 			$phpbb_userid = tep_db_fetch_array($phpbb_userid_query);
 			if ( true === empty($phpbb_customers_id['customers_user_id']) ){
	 			$update_user_id = "UPDATE " . TABLE_CUSTOMERS . " users SET customers_user_id = '{$phpbb_userid['username_clean']}' WHERE customers_id = '$customers_id'";
	 			tep_db_query($update_user_id);
 			}
	 		return $phpbb_userid['user_id'];
 		}
 	}
}


 function phpbb_user_logoff( $customers_userid ) {
 	$session_id = tep_session_id();
 	$phpbb_config_query_sql = "SELECT config_name, config_value FROM " . PHPBB_DB_PREFIX . "config";
 	$phpbb_config_query = tep_db_query($phpbb_config_query_sql);
 	if ( 0 < tep_db_num_rows($phpbb_config_query) ) {
 		while ( $phpbb_config = tep_db_fetch_array($phpbb_config_query) ) {
 			$config[$phpbb_config['config_name']] = $phpbb_config['config_value'];
 		}
 	}
 	header('Set-Cookie: ' . rawurlencode($config['cookie_name'] . '_sid') . '=; path=' . $config['cookie_path'] . $host . '; HttpOnly', false);
 	header('Set-Cookie: ' . rawurlencode($config['cookie_name'] . '_u') . '; path=' . $config['cookie_path'] . '; HttpOnly', false);
 	header('Set-Cookie: ' . rawurlencode($config['cookie_name'] . '_k') . '=; path=' . $config['cookie_path'] . '; HttpOnly', false);
 	$phpbb_session_delete = "DELETE FROM  " . PHPBB_DB_PREFIX . "sessions WHERE `session_id` = '$session_id'";
	tep_db_query($phpbb_session_delete);

 }

 function phpbb_account_is_active($phpbb_user_id) {
 	$active_query_sql = "SELECT user_inactive_reason FROM " . PHPBB_DB_PREFIX . "users WHERE user_id = '$phpbb_user_id'";
 	$active_query = tep_db_query($active_query_sql);
 	if ( 0 < tep_db_num_rows($active_query) ) {
 		$active = tep_db_fetch_array($active_query);
 		return ( '0' == $active['user_inactive_reason'] ? true : false );
 	} else {
 		return false;
 	}

 }


 function phpbb_user_login( $customers_id ) {
 	$phpbb_user_id = phpbb_get_user_id_by_customers_id($customers_id);
 	$time_now = time();
 	$browser = (!empty($_SERVER['HTTP_USER_AGENT'])) ? htmlspecialchars((string) $_SERVER['HTTP_USER_AGENT']) : '';
 	$return = array();
 	$return['error'] = false;
 	$return['message'] = '';

 	// Since HTTP_HOST may carry a port definition, we need to remove it here...
	if (strpos($host, ':') !== false) {
			$host = substr($host, 0, strpos($host, ':'));
	}
 	$session_id = tep_session_id();
 	$cookietime = $time_now+31536000;
 	$expire = gmdate('D, d-M-Y H:i:s \\G\\M\\T', $cookietime);
 	$phpbb_config_query_sql = "SELECT config_name, config_value FROM " . PHPBB_DB_PREFIX . "config";
 	$phpbb_config_query = tep_db_query($phpbb_config_query_sql);
 	$phpbb_account_is_active = phpbb_account_is_active($phpbb_user_id);
 	if ( false === $phpbb_account_is_active ) {
 		$return['message'] = TEXT_FORUM_ACCOUNT_INACTIVE;
 	}
 	if ( 0 < tep_db_num_rows($phpbb_config_query) && true === $phpbb_account_is_active ) {
 		while ( $phpbb_config = tep_db_fetch_array($phpbb_config_query) ) {
 			$config[$phpbb_config['config_name']] = $phpbb_config['config_value'];
 		}
 		$cookie_session_id = rawurlencode($config['cookie_name'] . '_sid') . '=' . rawurlencode($session_id);
 		$cookie_uid = rawurlencode($config['cookie_name'] . '_u') . '=' . rawurlencode($phpbb_user_id);
 		header('Set-Cookie: ' . $cookie_session_id . (($cookietime) ? '; expires=' . $expire : '') . '; path=' . $config['cookie_path'] . $host . '; HttpOnly', false);
 		header('Set-Cookie: ' . $cookie_uid . (($cookietime) ? '; expires=' . $expire : '') . '; path=' . $config['cookie_path'] . '; HttpOnly', false);
 		header('Set-Cookie: ' . rawurlencode($config['cookie_name'] . '_k') . '=' . (($cookietime) ? '; expires=' . $expire : '') . '; path=' . $config['cookie_path'] . '; HttpOnly', false);

 		$phpbb_session_check_sql = "SELECT * FROM " . PHPBB_DB_PREFIX . "sessions WHERE `session_id` = '$session_id'";
 		$phpbb_session_check = tep_db_query($phpbb_session_check_sql);
 		if ( 0 < tep_db_num_rows($phpbb_session_check) ) {
 			$phpbb_session_update = "UPDATE " . PHPBB_DB_PREFIX . "sessions
 							SET
 								`session_user_id` = '$phpbb_user_id',
 								`session_last_visit` = '$time_now',
 								`session_ip` = '{$_SERVER['REMOTE_ADDR']}',
 								`session_browser` = '$browser'
 							WHERE `session_id` = '$session_id'";
	 		tep_db_query($phpbb_session_update);
 		} else {
	 		$phpbb_session_insert = "INSERT INTO " . PHPBB_DB_PREFIX . "sessions
	 							(
	 								`session_id`,
	 								`session_user_id`,
	 								`session_last_visit`,
	 								`session_start`,
	 								`session_time`,
	 								`session_ip`,
	 								`session_browser`,
	 								`session_forwarded_for`,
	 								`session_page`,
	 								`session_autologin`,
	 								`session_admin`
	 							)
	 								VALUES
	 							(
	 								'$session_id',
	 								'$phpbb_user_id',
	 								'$time_now',
	 								'".$time_now."',
	 								'$time_now',
	 								'{$_SERVER['REMOTE_ADDR']}',
	 								'$browser',
	 								'',
	 								'index.php',
	 								1,
	 								0
	 							)";
	 		tep_db_query($phpbb_session_insert);
 		}
 	}
 	return $return;
}

function phpbb_create_account($email, $password){
	$check_account_query_sql = "SELECT customers_id FROM " . TABLE_CUSTOMERS . " WHERE `customers_email_address` = '$email'";
 	$check_account_query = tep_db_query($check_account_query_sql);
 	if ( 0 >= tep_db_num_rows($check_account_query) ) {
 		$phpbb_account_query_sql = "SELECT user_id, username_clean, user_password FROM " . PHPBB_DB_PREFIX . "users WHERE `user_email` = '$email'";
 		$phpbb_account_query = tep_db_query($phpbb_account_query_sql);
 		if ( 0 < tep_db_num_rows($phpbb_account_query) ) {
 			$phpbb_account = tep_db_fetch_array($phpbb_account_query);
 			$password_checked = phpbb_check_hash($password, $phpbb_account['user_password']);
 			if ( true === $password_checked ) {
	 			$password_hash = tep_encrypt_password($password);
	 			$insert_account_sql = "INSERT INTO " . TABLE_CUSTOMERS . "
	 								(
	 									customers_email_address,
	 									customers_user_id,
	 									customers_password
	 								)
	 									VALUES
	 								(
	 									'$email',
	 									'{$phpbb_account['username_clean']}',
	 									'$password_hash'
	 								) ";
	 			return tep_db_query($insert_account_sql);
	 		}
 		}
 	}
}

function phpbb_create_forum_account($email_address, $username, $password) {
	if ( true === empty($email_address) || true === empty($email_address) ) {
		return false;
	}
	$password_hashed = phpbb_hash($password);
	$user_email_hash = crc32(strtolower($email_address)) . strlen($email_address);
	$email_address = strtolower($email_address);
	$user_form_salt = unique_id();
	if ( true === defined('PHPBB_INTEGRATION') && 'Enabled' == PHPBB_INTEGRATION && true === defined('PHPBB_ACCOUNT_ACTIVATION') && ( 'User' === PHPBB_ACCOUNT_ACTIVATION || 'Admin' === PHPBB_ACCOUNT_ACTIVATION ) ) {
		$user_type = '1';
		$user_inactive_reason = '1';
		$user_inactive_time = time();
		$user_lastmark = '0';
		$user_actkey = tep_create_random_value(10);
		$user_actkey = strtoupper($user_actkey);
	} else {
		$user_type = '0';
		$user_inactive_reason = '0';
		$user_inactive_time = '0';
		$user_lastmark = time();
		$user_actkey = '';
	}
	$phpbb_user_query_sql = "INSERT INTO
									" . PHPBB_DB_PREFIX . "users
								(
									`user_type`,
									`group_id`,
									`user_ip`,
									`user_regdate`,
									`username`,
									`username_clean`,
									`user_password`,
									`user_passchg`,
									`user_email`,
									`user_email_hash`,
									`user_birthday`,
									`user_lastvisit`,
									`user_last_confirm_key`,
									`user_lang`,
									`user_inactive_reason`,
									`user_inactive_time`,
									`user_lastmark`,
									`user_style`,
									`user_options`,
									`user_dateformat`,
									`user_colour`,
									`user_form_salt`,
									`user_actkey`
								)
									VALUES
								(
									'$user_type',
									'2',
									'{$_SERVER['REMOTE_ADDR']}',
									'".time()."',
									'$username',
									'".strtolower($username)."',
									'$password_hashed',
									'".time()."',
									'$email_address',
									'$user_email_hash',
									'0',
									'".time()."',
									'0',
									'en',
									'$user_inactive_reason',
									'$user_inactive_time',
									'$user_lastmark',
									'1',
									'895',
									'D M d, Y g:i a',
									'9E8DA7',
									'$user_form_salt',
									'$user_actkey'
								)";
		$phpbb_user_query = tep_db_query($phpbb_user_query_sql);
		$phpbb_user_id = tep_db_insert_id();
		$phpbb_user_groups_query_sql = "INSERT INTO " . PHPBB_DB_PREFIX . "user_group (`group_id`, `user_id`, `user_pending`) VALUES ('2', '$phpbb_user_id', '0')";
		tep_db_query($phpbb_user_groups_query_sql);
}
?>